import os
import numpy as np
import pylab as plt
import matplotlib._color_data as mcd

from itertools import cycle

from .. import haven_utils as hu

COLORS = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"]
# COLORS += list(mcd.CSS4_COLORS)
COLORS += list(mcd.XKCD_COLORS)


def get_plot(
    exp_list,
    savedir_base,
    x_metric,
    y_metric,
    mode="line",
    filterby_list=None,
    title_list=None,
    legend_list=None,
    log_metric_list=None,
    figsize=None,
    avg_across=None,
    fig=None,
    axis=None,
    ylim=None,
    xlim=None,
    legend_fontsize=12,
    y_fontsize=14,
    x_fontsize=14,
    ytick_fontsize=12,
    xtick_fontsize=12,
    title_fontsize=18,
    legend_kwargs=None,
    map_title_list=tuple(),
    map_xlabel_list=tuple(),
    map_ylabel_list=dict(),
    bar_agg="min",
    verbose=True,
    show_legend=True,
    legend_format=None,
    title_format=None,
    cmap=None,
    show_ylabel=True,
    plot_confidence=True,
    x_cumsum=False,
    score_list_name="score_list.pkl",
    result_step=0,
    map_legend_list=dict(),
    xticks=None,
    random_marker=True,
    show_legend_key=True,
):
    """Plots the experiment list in a single figure.

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : str
        A directory where experiments are saved
    x_metric : str
        Specifies metric for the x-axis
    y_metric : str
        Specifies metric for the y-axis
    title_list : [type], optional
        [description], by default None
    legend_list : [type], optional
        [description], by default None
    meta_list : [type], optional
        [description], by default None
    log_metric_list : [type], optional
        [description], by default None
    figsize : tuple, optional
        [description], by default (8, 8)
    avg_metric : [type], optional
        [description], by default None
    axis : [type], optional
        [description], by default None
    ylim : [type], optional
        [description], by default None
    xlim : [type], optional
        [description], by default None
    legend_fontsize : [type], optional
        [description], by default None
    y_fontsize : [type], optional
        [description], by default None
    ytick_fontsize : [type], optional
        [description], by default None
    xtick_fontsize : [type], optional
        [description], by default None
    legend_kwargs : [type], optional
        [description], by default None
    title_format: [str], optional
        [description], formatting of the title, by default None
    cmap: [str], optional
        [description], specify colormap, by default None
    random_marker: [bool], optional
        [description], whether to choose markers at random

    Returns
    -------
    fig : [type]
        [description]
    axis : [type]
        [description]

    Examples
    --------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base,
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> hr.get_plot(exp_list, savedir_base=savedir_base, x_metric='epoch', y_metric='train_loss', legend_list=['model'])
    """
    exp_list, style_list = hu.filter_exp_list(
        exp_list, filterby_list=filterby_list, savedir_base=savedir_base, verbose=verbose, return_style_list=True
    )
    if legend_list is None:
        legend_list = hu.get_diff_hparam(exp_list)
    # if len(exp_list) == 0:
    if axis is None:
        fig, axis = plt.subplots(nrows=1, ncols=1, figsize=figsize)
    # default properties
    if title_list is not None:
        title = get_label(title_list, exp_list[0], format_str=title_format)
    else:
        title = ""

    ylabel = y_metric
    xlabel = x_metric

    # map properties
    for map_dict in map_title_list:
        if title in map_dict:
            title = map_dict[title]

    for map_dict in map_xlabel_list:
        if x_metric in map_dict:
            xlabel = map_dict[x_metric]

    for map_dict in map_ylabel_list:
        if y_metric in map_dict:
            ylabel = map_dict[y_metric]

    # set properties
    axis.set_title(title, fontsize=title_fontsize)
    if ylim is not None:
        axis.set_ylim(ylim)
    if xlim is not None:
        axis.set_xlim(xlim)

    if log_metric_list and y_metric in log_metric_list:
        axis.set_yscale("log")
        ylabel = ylabel + " (log)"

    if log_metric_list and x_metric in log_metric_list:
        axis.set_xscale("log")
        xlabel = xlabel + " (log)"

    if show_ylabel:
        axis.set_ylabel(ylabel, fontsize=y_fontsize)

    axis.set_xlabel(xlabel, fontsize=x_fontsize)

    axis.tick_params(axis="x", labelsize=xtick_fontsize)
    axis.tick_params(axis="y", labelsize=ytick_fontsize)

    axis.grid(True)

    # if len(exp_list) > 50:
    #     if verbose:
    #         raise ValueError('many experiments in one plot...use filterby_list to reduce them')
    # if cmap is not None or cmap is not '':
    #     plt.rcParams["axes.prop_cycle"] = get_cycle(cmap)
    if mode == "pretty_plot":
        tools.pretty_plot

    bar_count = 0
    visited_exp_ids = set()
    plot_idx = 0
    color_cycle = cycle(COLORS)
    for exp_dict, style_dict in zip(exp_list, style_list):
        exp_id = hu.hash_dict(exp_dict)
        if exp_id in visited_exp_ids:
            continue

        savedir = os.path.join(savedir_base, exp_id)
        score_list_fname = os.path.join(savedir, score_list_name)

        # skipt if it does not exist
        if not os.path.exists(score_list_fname):
            if verbose:
                print("%s: %s does not exist..." % (exp_id, score_list_name))
            continue

        # get result
        result_dict = get_result_dict(
            exp_dict,
            savedir_base,
            x_metric,
            y_metric,
            plot_confidence=plot_confidence,
            exp_list=exp_list,
            avg_across=avg_across,
            verbose=verbose,
            x_cumsum=x_cumsum,
            score_list_name=score_list_name,
            result_step=result_step,
        )
        # it is None if one of score_list.pkl is corrupted
        if result_dict is None:
            continue

        y_list = result_dict["y_list"]
        x_list = result_dict["x_list"]
        for eid in list(result_dict["visited_exp_ids"]):
            visited_exp_ids.add(eid)
        if len(x_list) == 0 or np.array(y_list).dtype == "object":
            x_list = np.NaN
            y_list = np.NaN
            if verbose:
                print('%s: "(%s, %s)" not in score_list' % (exp_id, y_metric, x_metric))

        # map properties of exp
        if legend_list is not None:
            if plot_idx != 0:
                show_legend_key = False
            else:
                show_legend_key = show_legend_key
            label = get_label(legend_list, exp_dict, format_str=legend_format, show_key=show_legend_key)
        else:
            label = exp_id

        color = next(color_cycle)
        plot_idx += 1

        if random_marker:
            marker = np.random.choice(
                ["o", "v", "<", ">", "1", "2", "3", "4", "p", "P", "s", "*", "+", "x", "X", "d", "D"]
            )
        else:
            marker = "o"

        linewidth = 2.8

        try:
            markevery = len(x_list) // 10
        except Exception:
            markevery = None
        if markevery == 0:
            markevery = None

        markersize = 6

        if len(style_dict):
            marker = style_dict.get("marker", marker)
            label = style_dict.get("label", label)
            color = style_dict.get("color", color)
            linewidth = style_dict.get("linewidth", linewidth)
            markevery = style_dict.get("markevery", markevery)
            markersize = style_dict.get("markersize", markersize)

        if label in map_legend_list:
            label = map_legend_list[label]
        # plot
        if mode == "pretty_plot":
            # plot the mean in a line
            # pplot = pp.add_yxList
            axis.plot(
                x_list,
                y_list,
                color=color,
                linewidth=linewidth,
                markersize=markersize,
                label=str(label),
                marker=marker,
                markevery=markevery,
            )
            # tools.pretty_plot
        elif mode == "line":
            # plot the mean in a line
            (line_plot,) = axis.plot(
                x_list,
                y_list,
                color=color,
                linewidth=linewidth,
                markersize=markersize,
                label=label,
                marker=marker,
                markevery=markevery,
            )

            if avg_across and hasattr(y_list, "size"):
                # add confidence interval
                axis.fill_between(
                    x_list,
                    y_list - result_dict.get("y_std_list", 0),
                    y_list + result_dict.get("y_std_list", 0),
                    color=line_plot.get_color(),
                    alpha=0.1,
                )

        elif mode == "bar":
            # plot the mean in a line
            if bar_agg == "max":
                y_agg = np.max(y_list)
            elif bar_agg == "min":
                y_agg = np.min(y_list)
            elif bar_agg == "mean":
                y_agg = np.mean(y_list)
            elif bar_agg == "last":
                y_agg = [y for y in y_list if isinstance(y, float)][-1]

            width = 0.0
            import math

            if math.isnan(y_agg):
                s = "NaN"
                continue

            else:
                s = "%.3f" % y_agg

            axis.bar(
                [bar_count + width],
                [y_agg],
                color=color,
                label=label,
                # label='%s - (%s: %d, %s: %.3f)' % (label, x_metric, x_list[-1], y_metric, y_agg)
            )
            if color is not None:
                bar_color = color
            else:
                bar_color = "black"

            # minimum, maximum = axis.get_ylim()
            # y_height = .05 * (maximum - minimum)

            # axis.text(bar_count, y_agg + .01, "%.3f"%y_agg, color=bar_color, fontweight='bold')
            axis.text(
                x=bar_count,
                y=y_agg * 1.01,
                s=s,
                fontdict=dict(fontsize=(y_fontsize or 12)),
                color="black",
                fontweight="bold",
            )
            axis.set_xticks([])
            bar_count += 1
        else:
            raise ValueError("mode %s does not exist. Options: (line, bar)" % mode)

    legend_kwargs = legend_kwargs or {"loc": 2, "bbox_to_anchor": (1.05, 1), "borderaxespad": 0.0, "ncol": 1}

    if mode == "pretty_plot":
        pass
    elif show_legend:
        axis.legend(fontsize=legend_fontsize, **legend_kwargs)

    plt.tight_layout()
    if xticks is not None:
        plt.xticks(*xticks)

    return fig, axis


def get_label(original_list, exp_dict, format_str=None, show_key=False):
    key_list = []
    label_list = []
    exp_dict_flat = hu.flatten_column(exp_dict, flatten_list=True)
    for i, k in enumerate(original_list):
        if k == "exp_id":
            sub_dict = str(hu.hash_dict(exp_dict))
        else:
            sub_dict = None
            if k in exp_dict_flat:
                sub_dict = exp_dict_flat[k]
            # depth_list = k.split(".")
            # sub_dict = exp_dict
            # for d in depth_list:
            #     if sub_dict is None or d not in sub_dict:
            #         sub_dict = None
            #         break
            #     sub_dict = sub_dict[d]

        if i < (len(original_list) - 1):
            if format_str:
                key_list += [f"{k}"]
                label_list += [f"{str(sub_dict):{int(len(k))}}"]
            else:
                key_list += [f"{k}|"]
                label_list += [f"{str(sub_dict):{int(len(k))}}|"]
        else:
            key_list += [f"{k}"]
            label_list += [f"{str(sub_dict)}"]

    if format_str:
        label = format_str.format(*label_list)
    else:
        label = " ".join(label_list)
        if show_key:
            label = " ".join(key_list) + "\n" + label

    # label = '\n'.join(wrap(label, 40))
    return label


def get_result_dict(
    exp_dict,
    savedir_base,
    x_metric,
    y_metric,
    exp_list=None,
    avg_across=False,
    verbose=False,
    plot_confidence=True,
    x_cumsum=False,
    score_list_name="score_list.pkl",
    result_step=0,
):
    visited_exp_ids = set()
    exp_id = hu.hash_dict(exp_dict)
    savedir = os.path.join(savedir_base, exp_id)
    score_list_fname = os.path.join(savedir, score_list_name)

    # get scores
    if not avg_across:
        # get score list
        try:
            score_list = hu.load_pkl(score_list_fname)
        except Exception:
            return None

        x_list = []
        y_list = []
        for score_dict in score_list:
            if x_metric in score_dict and y_metric in score_dict:
                x_list += [score_dict[x_metric]]
                y_list += [score_dict[y_metric]]
        y_std_list = []

    else:
        assert exp_list is not None, "exp_list must be passed"
        # average score list across an hparam

        filter_dict = {k: exp_dict[k] for k in exp_dict if k not in avg_across}
        exp_sublist = hu.filter_exp_list(
            exp_list, filterby_list=[filter_dict], savedir_base=savedir_base, verbose=verbose
        )

        def count(d):
            return sum([count(v) if isinstance(v, dict) else 1 for v in d.values()])

        n_values = count(filter_dict) + 1
        exp_sublist = [sub_dict for sub_dict in exp_sublist if n_values == count(sub_dict)]
        # get score list
        x_dict = {}

        uniques = np.unique([sub_dict[avg_across] for sub_dict in exp_sublist])
        # print(uniques, len(exp_sublist))
        assert len(exp_sublist) > 0
        assert len(uniques) == len(exp_sublist)
        for ek, sub_dict in enumerate(exp_sublist):
            sub_id = hu.hash_dict(sub_dict)
            sub_score_list_fname = os.path.join(savedir_base, sub_id, score_list_name)

            if not os.path.exists(sub_score_list_fname):
                if verbose:
                    print("%s: %s does not exist..." % (sub_id, score_list_name))
                continue

            visited_exp_ids.add(sub_id)

            try:
                sub_score_list = hu.load_pkl(sub_score_list_fname)
            except Exception:
                if verbose:
                    print("%s: %s is corrupt..." % (sub_id, score_list_name))
                return None

            for score_dict in sub_score_list:
                if x_metric in score_dict and y_metric in score_dict:
                    x_val = score_dict[x_metric]
                    if not x_val in x_dict:
                        x_dict[x_val] = {}

                    x_dict[x_val][ek] = score_dict[y_metric]
        # import ipdb; ipdb.set_trace()
        if len(x_dict) == 0 or isinstance(list(x_dict[x_val].values())[0], dict):
            x_list = []
            y_list = []
            y_std_list = []
        else:
            counts = {}
            for v in x_dict.values():
                for k in v:
                    if k not in counts:
                        counts[k] = 0
                    counts[k] += 1

            # interpolate
            xx_yy = {xx: list(yy.values()) for xx, yy in x_dict.items() if len(yy) == len(exp_sublist)}
            if len(np.unique([v in counts.values()])) == 1 and len(xx_yy) != len(sub_score_list):
                y_list_all = []
                x_list = np.array(list(x_dict.keys()))
                x_list.sort()
                for k in counts.keys():
                    xp_fp = {i: j[k] for i, j in x_dict.items() if k in j}
                    y_list_all += [np.interp(x_list, list(xp_fp.keys()), list(xp_fp.values()))]
                y_list_all = np.vstack(y_list_all).T
                # print()
            else:
                y_list_all = np.array(list(xx_yy.values()))
                x_list = np.array(list(xx_yy.keys()))

            if plot_confidence:
                y_std_list = np.std(y_list_all, axis=1)
            else:
                y_std_list = 0
            y_list = np.mean(y_list_all, axis=1)

    if x_cumsum:
        x_list = np.cumsum(x_list)

    if result_step == 0:
        return {"y_list": y_list, "x_list": x_list, "y_std_list": y_std_list, "visited_exp_ids": visited_exp_ids}
    else:
        return {
            "y_list": y_list[::result_step],
            "x_list": x_list[::result_step],
            "y_std_list": y_std_list,
            "visited_exp_ids": visited_exp_ids,
        }



import sys
import os
import pprint
import numpy as np
from haven.haven_results import plots_line as pl
from haven import haven_results as hr
from haven import haven_utils as hu
from haven.haven_jobs import slurm_manager as sm

import pprint
import pandas as pd


def get_latex_table(
    score_df,
    legend=None,
    metrics=None,
    filter_dict=dict(),
    map_row_dict_dict=dict(),
    map_col_dict=dict(),
    decimals=None,
    transposed=True,
    **kwargs
):
    columns = legend
    rows = metrics
    # break it
    dicts = score_df.T.to_dict()

    dicts_new = {}
    for i in dicts:
        exp_score_dict = dicts[i]
        legend_label = "+".join(
            [str(map_row_dict_dict.get(exp_score_dict[l], exp_score_dict[l])) for l in exp_score_dict if l in rows]
        )
        metric_scores = {}
        for k, v in exp_score_dict.items():
            if k in columns:
                if decimals is not None:
                    v = np.round(v, decimals)
                metric_scores[k] = v
        dicts_new[legend_label] = metric_scores

    df_new = pd.DataFrame(dicts_new)
    if transposed:
        df_new = df_new.T

    return df_new.to_latex(**kwargs)



import os
import tqdm
import shutil
from .. import haven_utils as hu
from .. import haven_jupyter as hj


def to_dropbox(exp_list, savedir_base, dropbox_path, access_token, zipname):
    """[summary]

    Parameters
    ----------
    exp_list : [type]
        [description]
    savedir_base : [type]
        [description]
    dropbox_path : [type]
        [description]
    access_token : [type]
        [description]
    """
    # zip files
    exp_id_list = [hu.hash_dict(exp_dict) for exp_dict in exp_list]
    src_fname = os.path.join(savedir_base, zipname)
    out_fname = os.path.join(dropbox_path, zipname)
    zipdir(exp_id_list, savedir_base, src_fname)

    upload_file_to_dropbox(src_fname, out_fname, access_token)
    print("saved: https://www.dropbox.com/home/%s" % out_fname)


def upload_file_to_dropbox(src_fname, out_fname, access_token):
    import dropbox

    dbx = dropbox.Dropbox(access_token)
    try:
        dbx.files_delete_v2(out_fname)
    except Exception:
        pass
    # with open(src_fname, 'rb') as f:
    #     dbx.files_upload(f.read(), out_fname)

    upload(access_token=access_token, file_path=src_fname, target_path=out_fname)


def upload(
    access_token,
    file_path,
    target_path,
    timeout=900,
    chunk_size=4 * 1024 * 1024,
):
    import os
    import dropbox
    import tqdm

    dbx = dropbox.Dropbox(access_token, timeout=timeout)
    with open(file_path, "rb") as f:
        file_size = os.path.getsize(file_path)
        chunk_size = 4 * 1024 * 1024
        if file_size <= chunk_size:
            print(dbx.files_upload(f.read(), target_path))
        else:
            with tqdm.tqdm(total=file_size, desc="Uploaded") as pbar:
                upload_session_start_result = dbx.files_upload_session_start(f.read(chunk_size))
                pbar.update(chunk_size)
                cursor = dropbox.files.UploadSessionCursor(
                    session_id=upload_session_start_result.session_id,
                    offset=f.tell(),
                )
                commit = dropbox.files.CommitInfo(path=target_path)
                while f.tell() < file_size:
                    if (file_size - f.tell()) <= chunk_size:
                        print(dbx.files_upload_session_finish(f.read(chunk_size), cursor, commit))
                    else:
                        dbx.files_upload_session_append(
                            f.read(chunk_size),
                            cursor.session_id,
                            cursor.offset,
                        )
                        cursor.offset = f.tell()
                    pbar.update(chunk_size)
    print("uploaded!")


def zipdir(
    exp_id_list,
    savedir_base,
    src_fname,
    add_jupyter=True,
    verbose=1,
    fname_list=None,
    dropbox_path="/shared",
    access_token=None,
):
    import zipfile

    zipf = zipfile.ZipFile(src_fname, "w", zipfile.ZIP_DEFLATED)

    # ziph is zipfile handle
    if add_jupyter:
        abs_path = os.path.join(savedir_base, "results.ipynb")
        hj.create_jupyter(
            fname=abs_path, savedir_base="results/", overwrite=False, print_url=False, create_notebook=True
        )

        rel_path = "results.ipynb"
        zipf.write(abs_path, rel_path)
        os.remove(abs_path)

    n_zipped = 0
    if verbose:
        tqdm_bar = tqdm.tqdm
    else:

        def tqdm_bar(x):
            return x

    fname_all = ["score_list.pkl", "exp_dict.json"]
    if isinstance(fname_list, list):
        fname_all += fname_list

    for exp_id in tqdm_bar(exp_id_list):
        if not os.path.isdir(os.path.join(savedir_base, exp_id)):
            continue

        for fname in fname_all:
            abs_path = os.path.join(savedir_base, exp_id, fname)
            rel_path = os.path.join("results", exp_id, fname)
            if os.path.exists(abs_path):
                zipf.write(abs_path, rel_path)

        n_zipped += 1

    zipf.close()
    if verbose:
        print("Zipped: %d/%d exps in %s" % (n_zipped, len(exp_id_list), src_fname))

    if access_token is not None and access_token != "":
        out_fname = os.path.join(dropbox_path, src_fname)
        upload_file_to_dropbox(src_fname, out_fname, access_token)
        print("saved: https://www.dropbox.com/home/%s" % out_fname)



import pylab as plt

plt.rcParams.update({"figure.max_open_warning": 0})

import copy
import os
import numpy as np
import pandas as pd

from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd
from .tables_scores import *
from .plots_line import *
from .latex_tables import get_latex_table
from . import images_fig


def load_score(savedir_base, exp_dict, metric):
    df = pd.DataFrame(hu.load_pkl(os.path.join(savedir_base, hu.hash_dict(exp_dict), "score_list.pkl")))
    if metric not in df:
        return None
    series = df[metric]
    return series[series.last_valid_index()]


class ResultManager:
    def __init__(
        self,
        savedir_base,
        exp_list=None,
        filterby_list=None,
        verbose=True,
        has_score_list=False,
        exp_groups=None,
        mode_key=None,
        exp_ids=None,
        save_history=False,
        score_list_name="score_list.pkl",
        account_id=None,
        job_scheduler="toolkit",
        topk_tuple=None,
    ):
        """[summary]

        Parameters
        ----------
        savedir_base : [type]
            A directory where experiments are saved
        exp_list : [type], optional
            [description], by default None
        filterby_list : [type], optional
            [description], by default None
        has_score_list : [type], optional
            [description], by default False

        Example
        -------
        >>> from haven import haven_results as hr
        >>> savedir_base='../results'
        >>> rm = hr.ResultManager(savedir_base=savedir_base,
                                filterby_list=[{'dataset':'mnist'}],
                                verbose=1)
        >>> for df in rm.get_score_df():
        >>>     display(df)
        >>> fig_list = rm.get_plot_all(y_metric_list=['train_loss', 'val_acc'],
                                    order='groups_by_metrics',
                                    x_metric='epoch',
                                    figsize=(15,6),
                                    title_list=['dataset'],
                                    legend_list=['model'])
        """

        # sanity checks
        assert os.path.exists(savedir_base), "%s does not exist" % savedir_base

        self.exp_groups = {}
        if exp_groups is not None:
            if isinstance(exp_groups, dict):
                # load from a dict
                self.exp_groups = exp_groups
            elif os.path.exists(exp_groups):
                # load from a file
                self.exp_groups = hu.load_py(exp_groups).EXP_GROUPS
            else:
                raise ValueError("%s does not exist..." % exp_groups)

        # rest
        self.score_list_name = score_list_name
        self.mode_key = mode_key
        self.has_score_list = has_score_list
        self.save_history = save_history
        self.account_id = account_id
        self.job_scheduler = job_scheduler
        # get exp _list

        if exp_ids is not None:
            assert exp_list is None, "settings exp_ids require exp_list=None"
            assert exp_groups is None, "settings exp_ids require exp_groups=None"
            exp_list = []
            for exp_id in exp_ids:
                exp_list += [hu.load_json(os.path.join(savedir_base, exp_id, "exp_dict.json"))]
        else:
            if exp_list is None:
                exp_list = hu.get_exp_list(savedir_base=savedir_base, verbose=verbose)
            else:
                exp_list = exp_list

            if len(exp_list) == 0:
                raise ValueError("exp_list is empty...")

        exp_list_with_scores = [
            e for e in exp_list if os.path.exists(os.path.join(savedir_base, hu.hash_dict(e), score_list_name))
        ]
        if has_score_list:
            exp_list = exp_list_with_scores

        self.savedir_base = savedir_base

        self.filterby_list = filterby_list
        self.verbose = verbose

        self.n_exp_all = len(exp_list)

        self.exp_list = hu.filter_exp_list(
            exp_list, filterby_list=filterby_list, savedir_base=savedir_base, verbose=verbose
        )

        if topk_tuple is not None:
            k, metric, desc = topk_tuple

            scores = []
            tmp_list = []
            for e in self.exp_list:
                score = load_score(savedir_base, e, metric)
                if score is None:
                    continue
                scores += [score]
                tmp_list += [e]
            sorted_ind = np.argsort(scores)
            if desc:
                sorted_ind = sorted_ind[::-1]
            sorted_ind = sorted_ind[:k]
            self.exp_list = list(np.array(tmp_list)[sorted_ind])

        self.exp_list_all = copy.deepcopy(self.exp_list)
        if mode_key:
            for exp_dict in exp_list:
                exp_dict[mode_key] = 1
            for exp_dict in self.exp_list_all:
                exp_dict[mode_key] = 1

        _, self.exp_params, self.score_keys = self.get_score_df(
            flatten_columns=True, show_meta=False, return_columns=True, show_max_min=False
        )
        self.exp_groups["all"] = copy.deepcopy(self.exp_list_all)

    def get_state_dict(self):
        pass

    def load_state_dict(self, state_dict):
        pass

    def get_savedir_list(self):
        return [os.path.join(self.savedir_base, hu.hash_dict(e_dict)) for e_dict in self.exp_list]

    def get_plot(self, groupby_list=None, savedir_plots=None, filterby_list=None, **kwargs):
        fig_list = []
        filterby_list = filterby_list or self.filterby_list
        exp_groups = hu.group_exp_list(self.exp_list, groupby_list)

        for i, exp_list in enumerate(exp_groups):
            fig, ax = get_plot(
                exp_list=exp_list,
                savedir_base=self.savedir_base,
                filterby_list=filterby_list,
                verbose=self.verbose,
                score_list_name=self.score_list_name,
                **kwargs
            )
            fig_list += [fig]

            # save image if given
            if savedir_plots != "" and savedir_plots is not None:
                os.makedirs(savedir_plots, exist_ok=True)
                save_fname = os.path.join(savedir_plots, "%d.png" % i)
                fig.savefig(save_fname, bbox_inches="tight")

        return fig_list

    def get_plot_all(
        self,
        y_metric_list,
        order="groups_by_metrics",
        groupby_list=None,
        ylim_list=None,
        xlim_list=None,
        savedir_plots=None,
        legend_last_row_only=False,
        show_legend_all=None,
        **kwargs
    ):
        """[summary]

        Parameters
        ----------
        y_metric_list : [type]
            [description]
        order : str, optional
            [description], by default 'groups_by_metrics'

        Returns
        -------
        [type]
            [description]

        """

        if order not in ["groups_by_metrics", "metrics_by_groups"]:
            raise ValueError(
                "%s order is not defined, choose between %s" % (order, ["groups_by_metrics", "metrics_by_groups"])
            )
        exp_groups = hu.group_exp_list(self.exp_list, groupby_list)
        figsize = kwargs.get("figsize") or None

        fig_list = []

        if not isinstance(y_metric_list, list):
            y_metric_list = [y_metric_list]

        if ylim_list is not None:
            assert isinstance(ylim_list[0], list), "ylim_list has to be lists of lists"
        if xlim_list is not None:
            assert isinstance(xlim_list[0], list), "xlim_list has to be lists of lists"

        if order == "groups_by_metrics":
            for j, exp_list in enumerate(exp_groups):
                fig, ax_list = plt.subplots(nrows=1, ncols=len(y_metric_list), figsize=figsize)
                if not hasattr(ax_list, "size"):
                    ax_list = [ax_list]
                for i, y_metric in enumerate(y_metric_list):
                    if i == (len(y_metric_list) - 1):
                        show_legend = True
                    else:
                        show_legend = False

                    ylim = None
                    xlim = None
                    if ylim_list is not None:
                        assert len(ylim_list) == len(exp_groups), "ylim_list has to have %d rows" % len(exp_groups)
                        assert len(ylim_list[0]) == len(y_metric_list), "ylim_list has to have %d cols" % len(
                            y_metric_list
                        )
                        ylim = ylim_list[j][i]
                    if xlim_list is not None:
                        assert len(xlim_list) == len(exp_groups), "xlim_list has to have %d rows" % len(exp_groups)
                        assert len(xlim_list[0]) == len(y_metric_list), "xlim_list has to have %d cols" % len(
                            y_metric_list
                        )
                        xlim = xlim_list[j][i]
                    if show_legend_all is not None:
                        show_legend = show_legend_all
                    fig, _ = get_plot(
                        exp_list=exp_list,
                        savedir_base=self.savedir_base,
                        y_metric=y_metric,
                        fig=fig,
                        axis=ax_list[i],
                        verbose=self.verbose,
                        filterby_list=self.filterby_list,
                        show_legend=show_legend,
                        ylim=ylim,
                        xlim=xlim,
                        score_list_name=self.score_list_name,
                        **kwargs
                    )
                fig_list += [fig]

        elif order == "metrics_by_groups":

            for j, y_metric in enumerate(y_metric_list):
                fig, ax_list = plt.subplots(nrows=1, ncols=len(exp_groups), figsize=figsize)
                if not hasattr(ax_list, "size"):
                    ax_list = [ax_list]
                for i, exp_list in enumerate(exp_groups):
                    if i == 0:
                        show_ylabel = True
                    else:
                        show_ylabel = False

                    if i == (len(exp_groups) - 1):
                        show_legend = True
                    else:
                        show_legend = False

                    if legend_last_row_only and j < (len(y_metric_list) - 1):
                        show_legend = False

                    ylim = None
                    xlim = None
                    if ylim_list is not None:
                        assert len(ylim_list) == len(y_metric_list), "ylim_list has to have %d rows" % len(exp_groups)
                        assert len(ylim_list[0]) == len(exp_groups), "ylim_list has to have %d cols" % len(
                            y_metric_list
                        )
                        ylim = ylim_list[j][i]
                    if xlim_list is not None:
                        assert len(xlim_list) == len(y_metric_list), "xlim_list has to have %d rows" % len(exp_groups)
                        assert len(xlim_list[0]) == len(exp_groups), "xlim_list has to have %d cols" % len(
                            y_metric_list
                        )
                        xlim = xlim_list[j][i]

                    if show_legend_all is not None:
                        show_legend = show_legend_all

                    fig, _ = get_plot(
                        exp_list=exp_list,
                        savedir_base=self.savedir_base,
                        y_metric=y_metric,
                        fig=fig,
                        axis=ax_list[i],
                        verbose=self.verbose,
                        filterby_list=self.filterby_list,
                        ylim=ylim,
                        xlim=xlim,
                        show_legend=show_legend,
                        show_ylabel=show_ylabel,
                        score_list_name=self.score_list_name,
                        **kwargs
                    )
                fig_list += [fig]

        plt.tight_layout()
        if savedir_plots:
            for i in range(len(fig_list)):
                os.makedirs(savedir_plots, exist_ok=True)
                fname = os.path.join(savedir_plots + "_%d.pdf" % i)
                fig_list[i].savefig(fname, dpi=300, bbox_inches="tight")
                print(fname, "saved")

        return fig_list

    def get_latex_table(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        score_df = self.get_score_df()
        df_list = get_latex_table(score_df=score_df, **kwargs)
        return df_list

    def get_score_df(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        df_list = get_score_df(
            exp_list=self.exp_list,
            savedir_base=self.savedir_base,
            verbose=self.verbose,
            score_list_name=self.score_list_name,
            **kwargs
        )
        return df_list

    def to_dropbox(self, outdir_base, access_token):
        """[summary]"""
        hd.to_dropbox(self.exp_list, savedir_base=self.savedir_base, outdir_base=outdir_base, access_token=access_token)

    def get_exp_list_df(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        df_list = get_exp_list_df(exp_list=self.exp_list, verbose=self.verbose, **kwargs)

        return df_list

    def get_exp_table(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        table = get_exp_list_df(exp_list=self.exp_list, verbose=self.verbose, **kwargs)
        return table

    def get_score_table(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        table = get_score_df(
            exp_list=self.exp_list,
            savedir_base=self.savedir_base,
            score_list_name=self.score_list_name,
            filterby_list=self.filterby_list,
            verbose=self.verbose,
            **kwargs
        )
        return table

    def get_score_lists(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        score_lists = get_score_lists(
            exp_list=self.exp_list,
            savedir_base=self.savedir_base,
            score_list_name=self.score_list_name,
            filterby_list=self.filterby_list,
            verbose=self.verbose,
            **kwargs
        )
        return score_lists

    def get_images(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        return images_fig.get_images(
            exp_list=self.exp_list, savedir_base=self.savedir_base, verbose=self.verbose, **kwargs
        )

    def get_job_summary(self, columns=None, add_prefix=False, **kwargs):
        """[summary]"""
        exp_list = hu.filter_exp_list(
            self.exp_list, self.filterby_list, savedir_base=self.savedir_base, verbose=self.verbose
        )
        jm = hjb.JobManager(
            exp_list=exp_list,
            savedir_base=self.savedir_base,
            account_id=self.account_id,
            job_scheduler=self.job_scheduler,
        )
        summary_list = jm.get_summary_list(columns=columns, add_prefix=add_prefix)

        return summary_list

    def kill_jobs(self, job_scheduler=None, **kwargs):
        """[summary]"""
        exp_list = hu.filter_exp_list(
            self.exp_list, self.filterby_list, savedir_base=self.savedir_base, verbose=self.verbose
        )
        jm = hjb.JobManager(
            exp_list=exp_list, savedir_base=self.savedir_base, account_id=self.account_id, job_scheduler=job_scheduler
        )

        return jm.kill_jobs()

    def to_zip(self, savedir_base="", fname="tmp.zip", **kwargs):
        """[summary]

        Parameters
        ----------
        fname : [type]
            [description]
        """
        if savedir_base == "":
            savedir_base = self.savedir_base
        exp_id_list = [hu.hash_dict(exp_dict) for exp_dict in self.exp_list]
        hd.zipdir(exp_id_list, savedir_base, fname, **kwargs)

    def to_dropbox(self, fname, dropbox_path=None, access_token=None):
        """[summary]

        Parameters
        ----------
        fname : [type]
            [description]
        dropbox_path : [type], optional
            [description], by default None
        access_token : [type], optional
            [description], by default None
        """
        from haven import haven_dropbox as hd

        out_fname = os.path.join(dropbox_path, fname)
        src_fname = os.path.join(self.savedir_base, fname)
        self.to_zip(src_fname)
        hd.upload_file_to_dropbox(src_fname, out_fname, access_token)
        print("saved: https://www.dropbox.com/home/%s" % out_fname)

    def plot_score_lists_from_exp_ids(
        self, score_list=None, score_lists_dict=None, x_metric="epoch", y_metric_list=("train_loss",), exp_ids=None
    ):
        if score_lists_dict is None:
            score_lists_dict = self.get_score_lists(return_as_dict=True)

        if exp_ids is None:
            exp_id_list = list(score_lists_dict.keys())
            exp_ids = []
            for i in exp_indices:
                exp_ids += [exp_id_list[i]]

        for exp_id in exp_ids:
            score_list = score_lists_dict[exp_id]
            plot_score_lists(score_list, x_metric, y_metric_list, exp_ids)

    def plot_score_lists(self, score_lists=None, x_metric="epoch", y_metric_list=("train_loss",), exp_ids=None):
        if exp_ids is not None:
            score_lists_dict = self.get_score_lists(return_as_dict=True)
            score_lists = []
            for ei in exp_ids:
                score_lists += [score_lists_dict[ei]]

        if score_lists is None:
            score_lists = self.get_score_lists()[:3]

        if not isinstance(score_lists[0], list):
            score_lists = [score_lists]

        for i, sc in enumerate(score_lists):
            df = pd.DataFrame(sc)
            df = df.fillna(method="bfill").fillna(method="ffill")

            y_list = []
            for y_metric in y_metric_list:
                if y_metric in df.columns:
                    y_list += [y_metric]

            df = df[[x_metric] + y_list]

            fig, ax = plt.subplots(figsize=(14, 7))
            y, x = df[x_metric].values, df.drop(x_metric, axis=1).T
            ax.stackplot(y, x, labels=x[0].keys())
            ax.legend(loc="upper left")
            if exp_ids is not None:
                plt.title(exp_ids[i])
            ax.set_xlabel(x_metric)
            plt.show()
